import 'dart:async';

import 'package:appflowy/workspace/application/workspace/workspace_listener.dart';
import 'package:appflowy/workspace/application/workspace/workspace_service.dart';
import 'package:appflowy_backend/log.dart';
import 'package:appflowy_backend/protobuf/flowy-error/errors.pb.dart';
import 'package:appflowy_backend/protobuf/flowy-folder/view.pb.dart';
import 'package:appflowy_backend/protobuf/flowy-user/user_profile.pb.dart';
import 'package:appflowy_result/appflowy_result.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:freezed_annotation/freezed_annotation.dart';

part 'menu_bloc.freezed.dart';

class MenuBloc extends Bloc<MenuEvent, MenuState> {
  MenuBloc({required this.user, required this.workspaceId})
      : _workspaceService = WorkspaceService(workspaceId: workspaceId),
        _listener = WorkspaceListener(
          user: user,
          workspaceId: workspaceId,
        ),
        super(MenuState.initial()) {
    _dispatch();
  }

  final WorkspaceService _workspaceService;
  final WorkspaceListener _listener;
  final UserProfilePB user;
  final String workspaceId;

  @override
  Future<void> close() async {
    await _listener.stop();
    return super.close();
  }

  void _dispatch() {
    on<MenuEvent>(
      (event, emit) async {
        await event.map(
          initial: (e) async {
            _listener.start(appsChanged: _handleAppsOrFail);
            await _fetchApps(emit);
          },
          createApp: (_CreateApp event) async {
            final result = await _workspaceService.createApp(
              name: event.name,
              desc: event.desc,
              index: event.index,
            );
            result.fold(
              (app) => emit(state.copyWith(lastCreatedView: app)),
              (error) {
                Log.error(error);
                emit(
                  state.copyWith(
                    successOrFailure: FlowyResult.failure(error),
                  ),
                );
              },
            );
          },
          didReceiveApps: (e) async {
            emit(
              e.appsOrFail.fold(
                (views) => state.copyWith(
                  views: views,
                  successOrFailure: FlowyResult.success(null),
                ),
                (err) =>
                    state.copyWith(successOrFailure: FlowyResult.failure(err)),
              ),
            );
          },
          moveApp: (_MoveApp value) {
            if (state.views.length > value.fromIndex) {
              final view = state.views[value.fromIndex];
              _workspaceService.moveApp(
                appId: view.id,
                fromIndex: value.fromIndex,
                toIndex: value.toIndex,
              );
              final apps = List<ViewPB>.from(state.views);

              apps.insert(value.toIndex, apps.removeAt(value.fromIndex));
              emit(state.copyWith(views: apps));
            }
          },
        );
      },
    );
  }

  // ignore: unused_element
  Future<void> _fetchApps(Emitter<MenuState> emit) async {
    final viewsOrError = await _workspaceService.getViews();
    emit(
      viewsOrError.fold(
        (views) => state.copyWith(views: views),
        (error) {
          Log.error(error);
          return state.copyWith(successOrFailure: FlowyResult.failure(error));
        },
      ),
    );
  }

  void _handleAppsOrFail(FlowyResult<List<ViewPB>, FlowyError> appsOrFail) {
    appsOrFail.fold(
      (apps) => add(MenuEvent.didReceiveApps(FlowyResult.success(apps))),
      (error) => add(MenuEvent.didReceiveApps(FlowyResult.failure(error))),
    );
  }
}

@freezed
class MenuEvent with _$MenuEvent {
  const factory MenuEvent.initial() = _Initial;
  const factory MenuEvent.createApp(String name, {String? desc, int? index}) =
      _CreateApp;
  const factory MenuEvent.moveApp(int fromIndex, int toIndex) = _MoveApp;
  const factory MenuEvent.didReceiveApps(
    FlowyResult<List<ViewPB>, FlowyError> appsOrFail,
  ) = _ReceiveApps;
}

@freezed
class MenuState with _$MenuState {
  const factory MenuState({
    required List<ViewPB> views,
    required FlowyResult<void, FlowyError> successOrFailure,
    ViewPB? lastCreatedView,
  }) = _MenuState;

  factory MenuState.initial() => MenuState(
        views: [],
        successOrFailure: FlowyResult.success(null),
      );
}
